import pytest
from werkzeug.serving import make_server
from threading import Thread
from tests.test_webapp.app import app


class ServerThread(Thread):
    def __init__(self, app):
        Thread.__init__(self)
        self.server = make_server("127.0.0.1", 5000, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        self.server.serve_forever()

    def shutdown(self):
        self.server.shutdown()


@pytest.fixture(scope="session")
def server():
    server = ServerThread(app)
    server.start()
    yield server
    server.shutdown()



"""The main parser for finding and checking links

The LinkResults class expects a URL as its only argument. It will return a 
dictionary of all links on the given html page which is stores in the object.results 
attribute. The returning dictionary consists of the link (dict value) and the http 
response (dict key).

    Typical usage example:

    results = LinkResults(url).results

The LinkAnalysis class expects a URL as its only argument, also. It returns a 
dictionary of all links that have the `target` attribute, and then analyzes to 
see if they are safe links (i.e. have a `rel` attribute, too). Links are also 
checked for obsolote attributes.

    Typical usage example:

    obsolete = LinkAnalysis(url).obsolete_attrs
    or
    unsafe = LinkAnalysis(url).unsafe_attrs
"""

import requests
from bs4 import BeautifulSoup
import re


class AllTags:
    def __init__(self, url: str):
        self.base_url = url if url.endswith("/") else url + "/"
        self.all_atags = self.find_all_atags(self.base_url)

    def __str__(self) -> str:
        return f"All links for {self.base_url}"

    def find_all_atags(self, url: str):
        """Find all anchor tags on a given URL.
        
        Args:
            url: A URL string

        Returns:
            A bs4.element.ResultSet object
        """

        r = requests.get(url)
        if r.status_code == 200:
            soup = BeautifulSoup(r.content, "html.parser")
            return soup.find_all("a")


class LinkResults(AllTags):
    def __init__(self, url: str):
        super().__init__(url)
        self.results = self.build_results_dictionary()

    def check_link_for_http_scheme(self, href: str) -> str:
        """Checks a link for http scheme.
        
        If a qualified URL, nothing is done. If a relative link, then a full URL 
        is built using the base URL.

        Args:
            href: The hypertext reference of a give anchor tag.

        Returns:
            A full qualifying URL
         """

        if href.startswith(self.base_url):
            return href

        elif href.startswith("/"):
            href = self.base_url + href.lstrip("/")
            return href

        elif href.startswith("./"):
            return self.base_url + href.lstrip("./")

        elif href.startswith("../"):
            return self.base_url + href.lstrip("../")

        else:
            return None  # Catches any links starting with hash(#)

    def build_results_dictionary(self) -> dict:
        """Build the final results dictionary.
        
        Once all_tags has been populated in the AllTags base class, the final 
        results dictionary is build by testing each link.

        Returns:
            A dictionary with the key = the HTTP response,
            and the value = a List of URLs that achieved that response code.
        """

        results = {}

        try:
            for tag in self.all_atags:
                href = tag.get("href")
                parsed_url = self.check_link_for_http_scheme(href)
                if parsed_url is not None:
                    parsed_url_status_code = requests.get(parsed_url).status_code

                    if parsed_url_status_code in results:
                        results[parsed_url_status_code].append(parsed_url)
                    else:
                        results[parsed_url_status_code] = [parsed_url]
        except:
            pass

        return results


class LinkAnalysis(AllTags):
    def __init__(self, url: str):
        super().__init__(url)
        self.obsolete_attrs = self.obsolete_attributes(self.all_atags)
        self.unsafe_attrs = self.unsafe_attributes(self.all_atags)

    def obsolete_attributes(self, links) -> dict:
        """ Check links for the presence of obsolete attributes
        
        Args:
            links: Takes a bs4 ResultSet which is returned from AllTags in self.all_atags

        Returns:
            Returns a dictionary with the key being the `href` of the link, 
            and the value being a `List` of obsolete attributes that the link has. 
            Any link with zero obsolete attributes is ingnored.

        """

        OBSOLETE_ATTRS = ("charset", "coords", "name", "rev", "shape")

        return_dict = {}
        for link in links:
            obs_link_attrs = [
                attribute for attribute in OBSOLETE_ATTRS if link.get(attribute)
            ]

            if obs_link_attrs:
                href = link["href"]
                if href in return_dict:
                    for _attr in obs_link_attrs:
                        return_dict[href].append(_attr)
                else:
                    return_dict[href] = obs_link_attrs

        return return_dict

    def unsafe_attributes(self, links) -> dict:
        """ Checks to see if links are unsafe.

        All links with the `target` attribute are checked for the presence/absence
        of a `rel="noopener"`. This should be present on all anchor tags with the
        `target` attribute to ensure the linked site does not have access to 
        window.opener property.
        
        For more details: https://developer.mozilla.org/en-US/docs/Web/HTML/Link_types/noopener 

        Args:
            links: Takes a bs4 ResultSet which is returned from AllTags in self.all_atags

        Returns:
            Returns a dictionary with the key being the full anchor tag checked, 
            and the value being a `Bool`. False = safe link, True = unsafe.
        """
        return_dict = {}
        for link in links:
            if link.get("target"):
                if link.get("rel") and "noopener" in link.get("rel"):
                    return_dict[link] = False
                else:
                    return_dict[link] = True

        return return_dict



import pytest

from typer.testing import CliRunner
from pyanchor.cli import app


runner = CliRunner()


@pytest.mark.usefixtures("server")
class TestCli:
    def test_exception_on_invalid_url_http_scheme(self):
        result = runner.invoke(app, ["google.com"])
        assert result.exit_code == 1

    def test_successful_result_prints(self):
        results = runner.invoke(app, ["http://127.0.0.1:5000/", "--verbose"])
        assert "[ 200 ] - http://127.0.0.1:5000/about" in results.output

    def test_failing_result_prints(self):
        results = runner.invoke(app, ["http://127.0.0.1:5000/"])
        assert "[ 500 ] - http://127.0.0.1:5000/500" in results.output

    def test_sitemap_successful_result_prints(self):
        results = runner.invoke(
            app, ["http://127.0.0.1:5000/sitemap.xml", "--sitemap", "--verbose"]
        )
        assert "[ 200 ] - http://127.0.0.1:5000/about" in results.output

    def test_sitemap_failing_result_prints(self):
        results = runner.invoke(app, ["http://127.0.0.1:5000/sitemap.xml", "--sitemap"])
        assert "[ 500 ] - http://127.0.0.1:5000/500" in results.output



import typer
import requests
from bs4 import BeautifulSoup
from pyanchor.link_checker import LinkResults


app = typer.Typer()


def print_results(links: dict, verbose: bool) -> int:
    """Simple utility function to print to terminal
    
    Args:
        Dictionary of links from LinkResults.results
    Returns:
        Two integers for the number of successful and unsuccessful links checked.
    """

    num_of_successful_links = 0
    num_of_failed_links = 0
    for http_code, url_list in links.items():
        if http_code == 200:
            for url in url_list:
                num_of_successful_links += 1
                if verbose:
                    typer.echo(typer.style(f"[ {http_code} ] - {url}", fg="green"))
        elif http_code == 500:
            for url in url_list:
                typer.echo(typer.style(f"[ {http_code} ] - {url}", fg="red"))
                num_of_failed_links += 1
        else:
            for url in url_list:
                typer.echo(typer.style(f"[ {http_code} ] - {url}", fg="yellow"))
                num_of_failed_links += 1

    return num_of_successful_links, num_of_failed_links


def print_totals(success: int, failed: int):
    """Prints results to terminal"""
    typer.echo("========================")
    typer.echo(f"TOTAL LINKS CHECKED: {success + failed}")
    typer.echo(f"FAILED: {failed}")


@app.command()
def main(
    url: str,
    sitemap: bool = typer.Option(
        False, "--sitemap", help="Use if the URL is a sitemap.xml link"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="By default all 200 responses will be hidden from the final output. Use verbose to view ALL results",
    ),
):
    """Check for broken links on any given web page. Pass in a sitemap URL to
    check all link on a given website.
    """

    if not url.startswith("http"):
        raise ValueError("Please provide a URL with a valid HTTP scheme")

    if sitemap:
        set_of_urls = set()
        results = {}

        r = requests.get(url)
        if r.status_code == 200:
            soup = BeautifulSoup(r.content, "lxml")
            sitemap_links = soup.find_all("loc")

            for sitemap_link in sitemap_links:
                set_of_urls.add(sitemap_link.text)

        all_results = []
        for _url in set_of_urls:
            link_results = LinkResults(_url).results
            if len(link_results) > 0:
                all_results.append(link_results)

        success_totals = 0
        failed_totals = 0
        for results_dict in all_results:
            _success, _failed = print_results(results_dict, verbose=verbose)
            success_totals += _success
            failed_totals += _failed

    else:
        results = LinkResults(url).results
        success_totals, failed_totals = print_results(results, verbose=verbose)

    print_totals(success_totals, failed_totals)


if __name__ == "__main__":
    app()



from flask import Flask, render_template


app = Flask(__name__)


@app.route("/")
def index():
    return render_template("index.html")


@app.route("/about/")
def about():
    sub_links = ["/link-1", "/link-2"]
    return render_template("success.html", name="About", links=sub_links)


@app.route("/about/link-1")
def about_link_1():
    return render_template("success.html", name="About Link 1")


@app.route("/about/link-2")
def about_link_2():
    return render_template("success.html", name="About Link 2")


@app.route("/contact")
def contact():
    return render_template("success.html", name="Contact")


@app.route("/rel")
def relative_link():
    return render_template("success.html", name="Rel")


@app.route("/rel2")
def relative_link2():
    return render_template("success.html", name="Rel2")


@app.route("/login")
def login():
    return render_template("error.html", status_code=401), 401


@app.route("/500")
def five_hundred():
    return render_template("error.html", status_code=500), 500


@app.route("/sitemap.xml")
def sitemap():
    return render_template("sitemap.xml")


if __name__ == "__main__":
    app.run()



import pytest

from pyanchor.link_checker import LinkAnalysis, LinkResults


@pytest.fixture
def example_LinkResults_object():
    return LinkResults("http://127.0.0.1:5000/")


@pytest.fixture
def example_LinkAnalysis_object():
    return LinkAnalysis("http://127.0.0.1:5000/")


class TestLinkResults:
    def test_results_is_dict(self, example_LinkResults_object):
        assert isinstance(example_LinkResults_object.results, dict)

    def test_link_result_class_str(self, example_LinkResults_object):
        assert str(example_LinkResults_object) == "All links for http://127.0.0.1:5000/"

    def test_successful_parse_relative_links(self, example_LinkResults_object):
        assert "http://127.0.0.1:5000/rel" in example_LinkResults_object.results[200]
        assert "http://127.0.0.1:5000/rel2" in example_LinkResults_object.results[200]


class TestLinkAnalysis:
    def test_obsolete_attrs_returns_dict(self, example_LinkAnalysis_object):
        assert isinstance(example_LinkAnalysis_object.obsolete_attrs, dict)

    def test_obsolete_attrs_on_tag(self, example_LinkAnalysis_object):
        attr_list = example_LinkAnalysis_object.obsolete_attrs
        assert attr_list["/about/link-1"][0] == "charset"
        assert attr_list["/about/link-2"][0] == "name"

    def test_unsafe_attrs_returns_dict(self, example_LinkAnalysis_object):
        assert isinstance(example_LinkAnalysis_object.unsafe_attrs, dict)
